Tracking Object and Session Changes with Events — SQLAlchemy 1.0 Documentation

Release: 1.0.18 | Release Date: July 24, 2017

# SQLAlchemy 1.0 Documentation

# Tracking Object and Session Changes with Events

SQLAlchemy features an extensive Event Listening system used throughout the Core and ORM. Within the ORM, there are a wide variety of event listener hooks, which are documented at an API level at ORM Events. This collection of events has grown over the years to include lots of very useful new events as well as some older events that aren’t as relevant as they once were. This section will attempt to introduce the major event hooks and when they might be used.

## Persistence Events

Probably the most widely used series of events are the “persistence” events, which correspond to the flush process. The flush is where all the decisions are made about pending changes to objects and are then emitted out to the database in the form of INSERT, UPDATE, and DELETE staetments.

### `before_flush()`

The `SessionEvents.before_flush()` hook is by far the most generally useful event to use when an application wants to ensure that additional persistence changes to the database are made when a flush proceeds. Use `SessionEvents.before_flush()` in order to operate upon objects to validate their state as well as to compose additional objects and references before they are persisted. Within this event, it is safe to manipulate the Session’s state, that is, new objects can be attached to it, objects can be deleted, and indivual attributes on objects can be changed freely, and these changes will be pulled into the flush process when the event hook completes.

The typical `SessionEvents.before_flush()` hook will be tasked with scanning the collections `Session.new`, `Session.dirty` and `Session.deleted` in order to look for objects where something will be happening.

For illustrations of `SessionEvents.before_flush()`, see examples such as Versioning with a History Table and Versioning using Temporal Rows.

### `after_flush()`

The `SessionEvents.after_flush()` hook is called after the SQL has been emitted for a flush process, but before the state of the objects that were flushed has been altered. That is, you can still inspect the `Session.new`, `Session.dirty` and `Session.deleted` collections to see what was just flushed, and you can also use history tracking features like the ones provided by `AttributeState` to see what changes were just persisted. In the `SessionEvents.after_flush()` event, additional SQL can be emitted to the database based on what’s observed to have changed.

### `after_flush_postexec()`

`SessionEvents.after_flush_postexec()` is called soon after `SessionEvents.after_flush()`, but is invoked after the state of the objects has been modified to account for the flush that just took place. The `Session.new`, `Session.dirty` and `Session.deleted` collections are normally completely empty here. Use `SessionEvents.after_flush_postexec()` to inspect the identity map for finalized objects and possibly emit additional SQL. In this hook, there is the ability to make new changes on objects, which means the `Session` will again go into a “dirty” state; the mechanics of the `Session` here will cause it to flush again if new changes are detected in this hook if the flush were invoked in the context of `Session.commit()`; otherwise, the pending changes will be bundled as part of the next normal flush. When the hook detects new changes within a `Session.commit()`, a counter ensures that an endless loop in this regard is stopped after 100 iterations, in the case that an `SessionEvents.after_flush_postexec()` hook continually adds new state to be flushed each time it is called.

### Mapper-level Events

In addition to the flush-level hooks, there is also a suite of hooks that are more fine-grained, in that they are called on a per-object basis and are broken out based on INSERT, UPDATE or DELETE. These are the mapper persistence hooks, and they too are very popular, however these events need to be approached more cautiously, as they proceed within the context of the flush process that is already ongoing; many operations are not safe to proceed here.

The events are:

- `MapperEvents.before_insert()`
- `MapperEvents.after_insert()`
- `MapperEvents.before_update()`
- `MapperEvents.after_update()`
- `MapperEvents.before_delete()`
- `MapperEvents.after_delete()`

Each event is passed the `Mapper`, the mapped object itself, and the `Connection` which is being used to emit an INSERT, UPDATE or DELETE statement. The appeal of these events is clear, in that if an application wants to tie some activity to when a specific type of object is persisted with an INSERT, the hook is very specific; unlike the `SessionEvents.before_flush()` event, there’s no need to search through collections like `Session.new` in order to find targets. However, the flush plan which represents the full list of every single INSERT, UPDATE, DELETE statement to be emitted has already been decided when these events are called, and no changes may be made at this stage. Therefore the only changes that are even possible to the given objects are upon attributes local to the object’s row. Any other change to the object or other objects will impact the state of the `Session`, which will fail to function properly.

Operations that are not supported within these mapper-level persistence events include:

- `Session.add()`
- `Session.delete()`
- Mapped collection append, add, remove, delete, discard, etc.
- Mapped relationship attribute set/del events, i.e. `someobject.related = someotherobject`

The reason the `Connection` is passed is that it is encouraged that simple SQL operations take place here, directly on the `Connection`, such as incrementing counters or inserting extra rows within log tables. When dealing with the `Connection`, it is expected that Core-level SQL operations will be used; e.g. those described in SQL Expression Language Tutorial.

There are also many per-object operations that don’t need to be handled within a flush event at all. The most common alternative is to simply establish additional state along with an object inside its `__init__()` method, such as creating additional objects that are to be associated with the new object. Using validators as described in Simple Validators is another approach; these functions can intercept changes to attributes and establish additional state changes on the target object in response to the attribute change. With both of these approaches, the object is in the correct state before it ever gets to the flush step.

## Object Lifecycle Events

Another use case for events is to track the lifecycle of objects. This refers to the states first introduced at Quickie Intro to Object States.

As of SQLAlchemy 1.0, there is no direct event interface for tracking of these states. Events that can be used at the moment to track the state of objects include:

- `InstanceEvents.init()`
- `InstanceEvents.load()`
- `SessionEvents.before_attach()`
- `SessionEvents.after_attach()`
- `SessionEvents.before_flush()` - by scanning the session’s collections
- `SessionEvents.after_flush()` - by scanning the session’s collections

SQLAlchemy 1.1 will introduce a comprehensive event system to track the object persistence states fully and unambiguously.

## Transaction Events

Transaction events allow an application to be notifed when transaction boundaries occur at the `Session` level as well as when the `Session` changes the transactional state on `Connection` objects.

- `SessionEvents.after_transaction_create()`, `SessionEvents.after_transaction_end()` - these events track the logical transaction scopes of the `Session` in a way that is not specific to individual database connections. These events are intended to help with integration of transaction-tracking systems such as `zope.sqlalchemy`. Use these events when the application needs to align some external scope with the transactional scope of the `Session`. These hooks mirror the “nested” transactional behavior of the `Session`, in that they track logical “subtransactions” as well as “nested” (e.g. SAVEPOINT) transactions.
- `SessionEvents.before_commit()`, `SessionEvents.after_commit()`, `SessionEvents.after_begin()`, `SessionEvents.after_rollback()`, `SessionEvents.after_soft_rollback()` - These events allow tracking of transaction events from the perspective of database connections. `SessionEvents.after_begin()` in particular is a per-connection event; a `Session` that maintains more than one connection will emit this event for each connection individually as those connections become used within the current transaction. The rollback and commit events then refer to when the DBAPI connections themselves have received rollback or commit instructions directly.

## Attribute Change Events

The attribute change events allow interception of when specific attributes on an object are modified. These events include `AttributeEvents.set()`, `AttributeEvents.append()`, and `AttributeEvents.remove()`. These events are extremely useful, particularly for per-object validation operations; however, it is often much more convenient to use a “validator” hook, which uses these hooks behind the scenes; see Simple Validators for background on this. The attribute events are also behind the mechanics of backreferences. An example illustrating use of attribute events is in Attribute Instrumentation.